﻿//console.debug('fdWebExtension - executing');

if (!window.fdConfigureExtension) {
  window.fdExtensionConfig = {
    enableLogging: false
  };

  window.fdConfigureExtension = function (config) {
    window.fdExtensionConfig.enableLogging = config.enableLogging;

    window.fdLog('fdConfigureExtension - enableLogging=' + window.fdExtensionConfig.enableLogging);
  }
}

if (!window.fdLog) {
  window.fdDisableLoggingForMessage = false;

  window.fdLog = function (logLine) {
    if (!window.fdExtensionConfig.enableLogging)
      return;
    else if (window.fdDisableLoggingForMessage)
      return;

    if (window.fdWebSocket) {
      window.fdWebSocket.send(JSON.stringify({
        id: window.fdGuid(),
        msg: 'WriteLogLine',
        logLine: logLine
      }));
    }
    else {
      console.debug(logLine);
    }
  }
}

// Create functions necessary for script to connect to and disconnect from FD.
// Also, create functions to support message handling.
if (!window.fdConnect) {
  window.fdConnect = function () {
    if (window.fdWebSocket) {
      //console.debug("fdConnect - already connected to FD");
      return;
    }

    //console.debug('fdConnect - connecting to FD');

    // only allow script to retry to connect to FD after fdConnect is called
    window.fdRetryConnectTimerId = null;
    window.fdRetryConnectEnabled = true;

    window.fdWebSocket = new WebSocket("wss://localhost:2020/");
    window.fdWebSocket.onopen = function () {
      //console.debug("fdWebSocket - opened");
    };
    window.fdWebSocket.onclose = function (evt) {
      //console.debug("fdWebSocket - closed: " + evt.code);
      window.fdWebSocket = null;

      // if FD isn't running yet or is restarted, then this here will ensure the script connects as soon as FD is running
      if (window.fdRetryConnectEnabled) {
        window.fdRetryConnectTimerId = window.setTimeout(fdConnect, 2000);
      }
    };
    window.fdWebSocket.onmessage = function (evt) {
      try {
        var obj = JSON.parse(evt.data);
        if (!obj || !obj.msg) {
          window.fdLog('fdWebSocket - unknown message format');
          return;
        }

        // FD polling code that sends messages will result in log spam
        // So give FD a way to disable logging per message
        if (obj.disableLogging) {
          window.fdDisableLoggingForMessage = true;
        }

        window.fdLog('fdWebSocket - handling msg=' + obj.msg);

        for (var i = 0; i < window.fdMessageHandlers.length; i++) {
          var startTime = Date.now();
          var handled = window.fdMessageHandlers[i].messageHandler(obj);
          if (handled) {
            var stopTime = Date.now();
            window.fdLog('fdWebSocket - finished handling msg=' + obj.msg + ' messageHandler=' + window.fdMessageHandlers[i].id + ' execTime=' + (stopTime - startTime) + 'ms');
            window.fdDisableLoggingForMessage = false;
            return;
          }
        }
      }
      catch (ex) {
        window.fdLog('fdWebSocket - ' + ex.stack);
      }

      window.fdLog("fdWebSocket - message was not handled");

      window.fdDisableLoggingForMessage = false;
    };
  };

  window.fdDisconnect = function () {
    // no longer allow script to retry to connect to FD
    window.fdRetryConnectEnabled = false;
    if (window.fdRetryConnectTimerId) {
      clearTimeout(window.fdRetryConnectTimerId);
    }

    if (!window.fdWebSocket)
      return;

    window.fdLog("fdDisconnect - closing WebSocket")

    window.fdWebSocket.close();
  }

  // messageHandler takes a message object and if the message is handled should return true else false
  // so if false is returned, then other message handlers will get the oppurtunity to handle the message
  window.fdMessageHandlers = [];
  window.fdAddMessageHandler = function (id, messageHandler) {
    window.fdMessageHandlers.push({
      id: id,
      messageHandler: messageHandler
    });
  };

  window.fdGuid = function () {
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
      var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
      return v.toString(16);
    });
  }

  window.fdAddMessageHandler('ConfigureExtensionHandler', function (obj) {
    if (obj.msg == 'ConfigureExtension') {
      window.fdConfigureExtension({
        enableLogging: obj.enableLogging
      });

      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: true,
          errorMessage: ''
        }
      }));

      return true;
    }

    return false;
  });
}

// Athena
// Create fdGetGrammar and fdExecuteCommand that wrap Athena's SpeechToText API
// that provides the ability to voice enable parts of the Athena UI.
// Create timer that polls Athena UI for contentEditable divs and sets role="text"
// attribute in order to get level 3 support.
if (window.AH) {
  window.fdLog('fdWebExtension - found Athena');

  if (window.SpeechToText) {
    window.fdLog('fdWebExtension - found Athena SpeechToText');

    window.fdGrammar = '';
    window.fdGrammarTimestamp = 0;

    if (!window.fdGetGrammar) {
      window.fdLog('fdWebExtension - defining Athena grammar functions');

      window.fdGetGrammarLogLevel = 0;
      window.fdGetGrammar = function () {
        if (!window.SpeechToText) {
          if (window.fdGetGrammarLogLevel == 0) {
            window.fdLog('fdGetGrammar - window.SpeechToText is null');
            window.fdGetGrammarLogLevel = 1;
          }
          window.fdGrammar = '';
          return window.fdGrammar;
        }

        if (!window.SpeechToText.V1_1) {
          if (window.fdGetGrammarLogLevel <= 1) {
            window.fdLog('fdGetGrammar - window.SpeechToText.V1_1 is null');
            window.fdGetGrammarLogLevel = 2;
          }
          window.fdGrammar = '';
          return window.fdGrammar;
        }

        var canRetrieveGrammar = window.SpeechToText.V1_1.CanRetrieveGrammar();
        if (!canRetrieveGrammar) {
          if (window.fdGetGrammarLogLevel <= 2) {
            window.fdLog('fdGetGrammar - CanRetrieveGrammar=false');
            window.fdGetGrammarLogLevel = 3;
          }
          window.fdGrammar = '';
          return window.fdGrammar;
        }

        var isGrammarReady = window.SpeechToText.V1_1.IsGrammarReady(window.fdGrammarTimestamp, 'JSGF');
        if (!isGrammarReady) {
          if (window.fdGetGrammarLogLevel <= 3) {
            window.fdLog('fdGetGrammar - IsGrammarReady(' + window.fdGrammarTimestamp + ')=false');
            window.fdGetGrammarLogLevel = 4;
          }
          window.fdGrammar = '';
          return window.fdGrammar;
        }

        var getGrammarStr = window.SpeechToText.V1_1.GetGrammar('JSGF');
        var getGrammarObj = JSON.parse(getGrammarStr);
        if (!getGrammarObj || !getGrammarObj.success || getGrammarObj.data === undefined || getGrammarObj.data.indexOf('not a supported grammar') > 0) {
          if (window.fdGetGrammarLogLevel <= 4) {
            window.fdLog('fdGetGrammar - GetGrammar failed - result=' + getGrammarStr);
            window.fdGetGrammarLogLevel = 5;
          }
          window.fdGrammar = '';
          return window.fdGrammar;
        }

        if (getGrammarObj.data) {
          window.fdGrammarTimestamp = getGrammarObj.timestamp;

          var grammar = getGrammarObj.data;
          grammar = grammar.replace(/(\r\n|\n|\r)/gm, ' ');
          grammar = grammar.replace(/ to/gm, " (to | two | too | 2)");
          grammar = grammar.replace(/ flowsheets/gm, " (flowsheets | (flow sheets))");
          grammar = grammar.replace(/ flowsheet/gm, " (flowsheet | (flow sheet))");
          grammar = grammar.replace(/ facesheet/gm, " (facesheet | (face sheet))");

          window.fdLog('fdGetGrammar - grammar=' + grammar);
          window.fdGetGrammarLogLevel = 0;  //reset function's log level so errors can be logged again

          window.fdGrammar = grammar;
          return window.fdGrammar;
        }
        else {
          // GetGrammar was successful, but grammar is empty
          // so use current grammar
          return window.fdGrammar;
        }
      }

      window.fdExecuteCommand = function (resultDetail) {
        window.fdLog("fdExecuteCommand - value='" + resultDetail.$value + "'");

        var resultValue = resultDetail.$value;
        resultValue = resultValue.replace(/ 2/gm, " to");
        resultValue = resultValue.replace(/ too/gm, " to");
        resultValue = resultValue.replace(/ two/gm, " to");
        resultValue = resultValue.replace(/ flow sheets/gm, " flowsheets");
        resultValue = resultValue.replace(/ flow sheet/gm, " flowsheet");
        resultValue = resultValue.replace(/ face sheet/gm, " facesheet");
        resultValue = resultValue.replace(/ tablet/gm, " tab");
        resultValue = resultValue.replace(/ tablets/gm, " tab");

        window.SpeechToText.V1_1.Enter(resultValue);
      }
    }
  }

  if (!window.fdUpdateContentEditableDiv) {
    window.fdUpdateContentEditableDiv = function (divEl) {
      if (!divEl.isContentEditable) {
        return;
      }

      var role = divEl.getAttribute('role');
      if (role && role == 'textbox') {
        return;
      }

      window.fdLog('fdUpdateContentEditableDiv - setting role=textbox on contentEditable div id=' + divEl.id);

      divEl.setAttribute('role', 'textbox');
    }
  }

  if (!window.fdUpdateContentEditableDivsInFrame) {
    window.fdUpdateContentEditableDivsInFrame = function (frame) {
      var divEls = frame.document.getElementsByTagName('div');
      for (var i = 0; i < divEls.length; i++) {
        var divEl = divEls[i];
        window.fdUpdateContentEditableDiv(divEl);
      }

      for (var i = 0; i < frame.frames.length; i++) {
        window.fdUpdateContentEditableDivsInFrame(frame.frames[i]);
      }
    }
  }

  if (!window.fdAthenaTimerId) {
    if (window.AH.Frames.Main) {
      window.fdLog('fdWebExtension - found Athena main frame, creating Athena timer');

      window.fdAthenaTimerId = window.setInterval(function (athenaMainFrame) {
        if (!athenaMainFrame.frame) {
          window.fdLog('fdAthenaTimer - Athena main frame is not loaded yet');
          return;
        }

        window.fdUpdateContentEditableDivsInFrame(athenaMainFrame.frame);
      }, 100, window.AH.Frames.Main);
    }
  }
}

if (!window.fdGetFocusedControlIdStandard) {
  window.fdLastUniqueControlIdNum = 0;
  window.fdGenerateUniqueControlId = function() {
    return "fdControl" + ++window.fdLastUniqueControlIdNum;
  }

  window.fdGetMainWindow = function (frame) {
    var mainWindow = frame ? frame.contentWindow : window;

    if (!mainWindow.document.activeElement) {
      return mainWindow;
    }
    else if (!mainWindow.document.activeElement.contentDocument) {
      return mainWindow;
    }
    else {
      return window.fdGetMainWindow(mainWindow.document.activeElement);
    }
  }

  window.fdIsTextInputOrTextArea = function (el) {
    return el.tagName == 'TEXTAREA' || (el.tagName == 'INPUT' && (el.type == 'text' || el.type == 'search'));
  }

  window.fdIsContentEditableEl = function (el) {
    return el.isContentEditable;
  }

  window.fdGetFocusedControlIdStandard = function () {
    var mainWindow = window.fdGetMainWindow();
    var activeElement = mainWindow.document.activeElement;
    if (!activeElement) {
      window.fdLog('fdGetFocusedControlIdStandard - no active element');
      return {
        success: true,
        focusedControlId: null
      };
    }

    var isStandardTextControl = window.fdIsTextInputOrTextArea(activeElement) || window.fdIsContentEditableEl(activeElement);
    if (!isStandardTextControl) {
      window.fdLog("fdGetFocusedControlIdStandard - id='" + activeElement.id + "' tagName='" + activeElement.tagName +
        "' isContentEditable=" + activeElement.isContentEditable + " is not a text control");

      return {
        success: true,
        focusedControlId: null
      };
    }

    if (!activeElement.fdUniqueId) {
      activeElement.fdUniqueId = window.fdGenerateUniqueControlId();
      activeElement.className += ' ' + activeElement.fdUniqueId;
    }

    window.fdLog("fdGetFocusedControlIdStandard - id='" + activeElement.id + "' fdUniqueId='" + activeElement.fdUniqueId +
      "' tagName='" + activeElement.tagName + "' isContentEditable=" + activeElement.isContentEditable);

    return {
      success: true,
      focusedControlId: activeElement.fdUniqueId
    };
  }

  window.fdGetContentEditableChildNodesInfo = function (childNodes, info, indentLogMsgSpacesCt) {
    if (!info.sel) {
      var mainWindow = window.fdGetMainWindow();
      var sel = mainWindow.getSelection();

      info.sel = sel;
      if (sel.isCollapsed) {
        info.isSelLeftToRight = true;
      }
      else if (sel.anchorNode.isSameNode(sel.focusNode)) {
        info.isSelLeftToRight = sel.focusOffset > sel.anchorOffset;
      }
      else if(childNodes[0].parentNode.isSameNode(sel.anchorNode) || childNodes[0].parentNode.isSameNode(sel.focusNode)) {
        info.isSelLeftToRight = true;
      }
      else {
        info.isSelLeftToRight = (sel.anchorNode.compareDocumentPosition(sel.focusNode) & Node.DOCUMENT_POSITION_FOLLOWING) > 0;
      }

      info.text = '';
      info.selStart = 0
      info.selEnd = 0;
      info.foundSelStart = false;
      info.foundSelEnd = false;
      info.currNodeStart = 0;
      info.currNodeEnd = 0;
      info.nodeList = [];
    }

    if (childNodes.length == 0) {
      return;
    }

    for (var i = 0; i < childNodes.length; i++) {
      var childNode = childNodes[i];

      if (childNode.parentNode.isSameNode(info.sel.anchorNode) && childNode.parentNode.isSameNode(info.sel.focusNode)) {
        if (i === info.sel.anchorOffset) {
          info.foundSelStart = true;
          info.foundSelEnd = true;
        }
      }
      else if (childNode.parentNode.isSameNode(info.sel.anchorNode)) {
        if (i === info.sel.anchorOffset) {
          if (info.foundSelEnd)
            info.isSelLeftToRight = false;
          info.foundSelStart = true;
        }
      }
      else if (childNode.parentNode.isSameNode(info.sel.focusNode)) {
        if (i === info.sel.focusOffset) {
          info.foundSelEnd = true;
        }
      }
      
      var isAnchorNode = childNode.isSameNode(info.sel.anchorNode);
      var isFocusNode = childNode.isSameNode(info.sel.focusNode);

      var text = '';
      if (childNode.nodeName == '#text') {
        text = childNode.nodeValue;
      }
      else if (childNode.nodeName == 'DIV' && childNode.childNodes.length == 1 && childNode.childNodes[0].nodeName == '#text') {
        text = '\n';
      }
      else if (childNode.nodeName == 'BR') {
        text = '\n';
      }

      if (text) {
        info.text += text;

        info.nodeList.push({
          node: childNode,
          text: text,
          start: info.currNodeEnd,
          end: info.currNodeEnd + text.length
        });
        info.currNodeStart = info.nodeList[info.nodeList.length - 1].start;
        info.currNodeEnd = info.nodeList[info.nodeList.length - 1].end;
      }

      // selection anchor and focus for empty lines is the outer div rather than inner line break
      // this throws off our selection start and/or end calculations
      // so detect the condition here and update selection start and/or end below by adding 1
      var isBrOnlyDiv = childNode.nodeName == 'DIV' && childNode.childNodes.length == 1 && childNode.childNodes[0].nodeName == 'BR';

      if (!info.foundSelStart) {
        if (info.isSelLeftToRight && isAnchorNode) {
          info.selStart += info.sel.anchorOffset;
          if (isBrOnlyDiv) {
            info.selStart++;
          }
          info.foundSelStart = true;
        }
        else if (!info.isSelLeftToRight && isFocusNode) {
          info.selStart += info.sel.focusOffset;
          if (isBrOnlyDiv) {
            info.selStart++;
          }
          info.foundSelStart = true;
        }
        else {
          info.selStart += text.length;
        }
      }

      if (!info.foundSelEnd) {
        if (info.isSelLeftToRight && isFocusNode) {
          info.selEnd += info.sel.focusOffset;
          if (isBrOnlyDiv) {
            info.selEnd++;
          }
          info.foundSelEnd = true;
        }
        else if (!info.isSelLeftToRight && isAnchorNode) {
          info.selEnd += info.sel.anchorOffset;
          if (isBrOnlyDiv) {
            info.selEnd++;
          }
          info.foundSelEnd = true;
        }
        else {
          info.selEnd += text.length;
        }
      }

      //WARNING - if there is a lot of text, then logging each node info causes FD to slow down a lot.
      //window.fdLog("fdGetContentEditableChildNodesInfo -" + new Array(indentLogMsgSpacesCt).join(' ') +
      //  " nodeName='" + childNode.nodeName + "' nodeValue='" + childNode.nodeValue + "' text='" + text.replace(/\n/g, '\\n') +
      //  "' isAnchorNode=" + isAnchorNode + " anchorOffset=" + (isAnchorNode ? info.sel.anchorOffset : 0) +
      //  " isFocusNode=" + isFocusNode + " focusOffset=" + (isFocusNode ? info.sel.focusOffset : 0) +
      //  " start=" + (text ? info.currNodeStart : -1) + ", end=" + (text ? info.currNodeEnd : -1));

      if(childNode.nodeName !== 'BR' && childNode.nodeName !== '#text'){
        window.fdGetContentEditableChildNodesInfo(childNode.childNodes, info, indentLogMsgSpacesCt + 2);
      }
    }
  }

  window.fdFindControlStandard = function (controlId) {
    window.fdLog('fdFindControlStandard - controlId=' + controlId);

    var mainWindow = window.fdGetMainWindow();
    var document = mainWindow.document;
    var textControlEl = document.querySelector('.' + controlId);

    if (textControlEl) {
      window.fdLog('fdFindControlStandard - found controlId=' + controlId)
    }

    return textControlEl;
  }

  window.fdGetTextStandard = function (textControlEl) {
    window.fdLog('fdGetTextStandard - controlId=' + textControlEl.fdUniqueId);

    var text = '';
    if (window.fdIsTextInputOrTextArea(textControlEl)) {
      text = textControlEl.value;
    }
    else {
      var info = {};
      window.fdGetContentEditableChildNodesInfo(textControlEl.childNodes, info, 0);
      text = info.text;
    }

    window.fdLog("fdGetTextStandard - controlId=" + textControlEl.fdUniqueId + " text='" + text.replace(/\n/g, '\\n') + "'");
      
    return {
      success: true,
      text: text
    }
  }

  window.fdGetTextLengthStandard = function (textControlEl) {
    window.fdLog('fdGetTextLengthStandard - controlId=' + textControlEl.fdUniqueId);

    var text = window.fdGetTextStandard(textControlEl).text;
    var textLength = text ? text.length : 0;

    window.fdLog('fdGetTextLengthStandard - controlId=' + textControlEl.fdUniqueId + ' textLength=' + textLength);

    return {
      success: true,
      textLength: textLength
    }
  }

  window.fdGetSelectionStandard = function (textControlEl) {
    window.fdLog('fdGetSelectionStandard - controlId=' + textControlEl.fdUniqueId);

    var selStart = 0;
    var selEnd = 0;

    if (window.fdIsTextInputOrTextArea(textControlEl)) {
      selStart = textControlEl.selectionStart;
      selEnd = textControlEl.selectionEnd;
    }
    else {
      var info = {};
      window.fdGetContentEditableChildNodesInfo(textControlEl.childNodes, info, 0);
      if (!info.isSelLeftToRight && info.selStart > info.selEnd) {
        selStart = info.selEnd;
        selEnd = info.selStart;
      } else {
        selStart = info.selStart;
        selEnd = info.selEnd;
      }
    }

    window.fdLog('fdGetSelectionStandard - controlId=' + textControlEl.fdUniqueId + ' selStart=' + selStart + ' selEnd=' + selEnd);

    return {
      success: true,
      start: selStart,
      end: selEnd,
      length: selEnd - selStart
    }
  }

  window.fdSetSelectionStandard = function (textControlEl, selStart, selEnd) {
    window.fdLog('fdSetSelectionStandard - controlId=' + textControlEl.fdUniqueId + ' selStart=' + selStart + ' selEnd=' + selEnd);

    if (window.fdIsTextInputOrTextArea(textControlEl)) {
      if (selEnd < 0) {
        window.fdLog('fdSetSelectionStandard - select all text');
        selEnd = textControlEl.value.length;
      }

      textControlEl.setSelectionRange(selStart, selEnd, 'forward');
    }
    else {
      var info = {};
      window.fdGetContentEditableChildNodesInfo(textControlEl.childNodes, info, 0);

      if (selEnd < 0 && info.nodeList.length > 0) {
        window.fdLog('fdSetSelectionStandard - select all text');
        selEnd = info.nodeList[info.nodeList.length - 1].end;
      }

      var startNode = null;
      var endNode = null;
      for (var i = 0; i < info.nodeList.length; i++) {
        var nodeStart = info.nodeList[i].start;
        var nodeEnd = info.nodeList[i].end;
        var isLastNode = i == info.nodeList.length - 1;
        var isTextNode = info.nodeList[i].node.nodeName == '#text';

        if (selStart >= nodeStart && selStart <= nodeEnd) {
          if (selStart == nodeEnd && !isLastNode && !isTextNode) {
            // set selection to start of next node instead of end of this node
            continue;
          }
          
          startNode = info.nodeList[i];
        }
        if (selEnd >= info.nodeList[i].start && selEnd <= info.nodeList[i].end) {
          if (selEnd == nodeEnd && !isLastNode && !isTextNode) {
            // set selection to start of next node instead of end of this node
            continue;
          }

          endNode = info.nodeList[i];
        }

        if (startNode && endNode) {
          break;
        }
      }

      if (startNode && endNode) {
        var mainWindow = window.fdGetMainWindow();
        var document = mainWindow.document;
        var range = document.createRange();
        if (!Boolean(window.chrome) && startNode === endNode && startNode.node.nodeName === "BR") {
          range.setStartAfter(endNode.node);
          range.setEndAfter(endNode.node);
        } else {
          if (startNode.node.nodeName === "BR")
            selStart = startNode.start;
          range.setStart(startNode.node, selStart - startNode.start);
          if (endNode.node.nodeName === "BR")
            selEnd = endNode.start;
          range.setEnd(endNode.node, selEnd - endNode.start);
        }
        var sel = document.getSelection();
        sel.removeAllRanges();
        sel.addRange(range);
      }
    }

    return { success: true };
  }

  window.fdInsertTextStandard = function (textControlEl, text) {
    window.fdLog("fdInsertTextStandard - controlId=" + textControlEl.fdUniqueId + " text='" + text.replace(/\n/g, '\\n') + "'");

    var mainWindow = window.fdGetMainWindow();
    var document = mainWindow.document;

    var oldScrollHeight = textControlEl.scrollHeight;
    var oldScrollTop = textControlEl.scrollTop;

    if (window.fdIsContentEditableEl(textControlEl) || Boolean(window.chrome)) {
      try {
        document.execCommand('insertText', false, text);
      } catch (e) { }
    }
    else {
      var selStart = textControlEl.selectionStart;
      var selEnd = textControlEl.selectionEnd;
      textControlEl.setRangeText(text, selStart, selEnd, "end");        
    }

    try
    {
      // simulate fake key down event to trigger application event handlers
      // works in Athena where it's necessary for text controls to grow
      // probably needed in more applications, so let's always simulate it
      var eventObj = document.createEventObject ? document.createEventObject() : document.createEvent("Events");
      if (eventObj.initEvent) {
        eventObj.initEvent("keydown", true, true);
      }
      textControlEl.dispatchEvent ? textControlEl.dispatchEvent(eventObj) : textControlEl.fireEvent("onkeydown", eventObj);
    }
    catch (e) { }

    var newScrollHeight = textControlEl.scrollHeight;
    var newScrollTop = textControlEl.scrollTop;

    if (newScrollHeight > oldScrollHeight && newScrollTop == oldScrollTop) {
      // if scrollable height increased due to insertion of text, then it must have been put on a new line
      // if scroll top hasn't changed, then new line wasn't scrolled into view
      // so do it now
      textControlEl.scrollTop = textControlEl.scrollTop + (newScrollHeight - oldScrollHeight);
    }

    return { success: true };
  }

  window.fdDeleteSelectionStandard = function (textControlEl) {
    window.fdLog("fdDeleteSelectionStandard - controlId=" + textControlEl.fdUniqueId);
    var mainWindow = window.fdGetMainWindow();

    if (window.fdIsContentEditableEl(textControlEl) || Boolean(window.chrome)) {
      mainWindow.document.execCommand('delete', false, null);
    }
    else {
        var selStart = textControlEl.selectionStart;
        var selEnd = textControlEl.selectionEnd;
        textControlEl.setRangeText('', selStart, selEnd, "end");
    }
    return { success: true };
  }

  window.fdBoldSelectionStandard = function (textControlEl) {
    window.fdLog("fdBoldSelectionStandard - controlId=" + textControlEl.fdUniqueId);

    var mainWindow = window.fdGetMainWindow();
    if (mainWindow.document.execCommand('bold', false, null)) {
      return {
        success: true,
        supported: true,
        sendKeysFromFD: false
      };
    }

    window.fdLog("fdBoldSelectionStandard - cannot bold text using API");

    var isTextInputOrTextArea = window.fdIsTextInputOrTextArea(textControlEl);
    return {
      success: false,
      errorMessage: 'cannot bold text using API',
      supported: !isTextInputOrTextArea,
      sendKeysFromFD: !isTextInputOrTextArea
    };
  }

  window.fdItalicizeSelectionStandard = function (textControlEl) {
    window.fdLog("fdItalicizeSelectionStandard - controlId=" + textControlEl.fdUniqueId);

    var mainWindow = window.fdGetMainWindow();
    if (mainWindow.document.execCommand('italicize', false, null)) {
      return {
        success: true,
        supported: true,
        sendKeysFromFD: false
      };
    }

    window.fdLog("fdItalicizeSelectionStandard - cannot italicize text using API");
    var isTextInputOrTextArea = window.fdIsTextInputOrTextArea(textControlEl);
    return {
      success: false,
      errorMessage: 'cannot italicize text using API',
      supported: !isTextInputOrTextArea,
      sendKeysFromFD: !isTextInputOrTextArea
    };
  }

  window.fdUnderlineSelectionStandard = function (textControlEl) {
    window.fdLog("fdUnderlineSelectionStandard - controlId=" + textControlEl.fdUniqueId);

    var mainWindow = window.fdGetMainWindow();
    if (mainWindow.document.execCommand('underline', false, null)) {
      return {
        success: true,
        supported: true,
        sendKeysFromFD: false
      };
    }

    window.fdLog("fdUnderlineSelectionStandard - cannot underline text using API");
    var isTextInputOrTextArea = window.fdIsTextInputOrTextArea(textControlEl);
    return {
      success: false,
      errorMessage: 'cannot underline text using API',
      supported: !isTextInputOrTextArea,
      sendKeysFromFD: !isTextInputOrTextArea
    };
  }

  window.fdCutSelectionStandard = function (textControlEl) {
    window.fdLog("fdCutSelectionStandard - controlId=" + textControlEl.fdUniqueId);

    var mainWindow = window.fdGetMainWindow();
    if (mainWindow.document.execCommand('cut', false, null)) {
      return {
        success: true,
        supported: true,
        sendKeysFromFD: false
      };
    }

    window.fdLog("fdCutSelectionStandard - cannot cut text using API");
    return {
      success: false,
      errorMessage: 'cannot cut text using API',
      supported: true,
      sendKeysFromFD: true
    };
  }

  window.fdCopySelectionStandard = function (textControlEl) {
    window.fdLog("fdCopySelectionStandard - controlId=" + textControlEl.fdUniqueId);

    var mainWindow = window.fdGetMainWindow();
    if (mainWindow.document.execCommand('copy', false, null)) {
      return {
        success: true,
        supported: true,
        sendKeysFromFD: false
      };
    }

    window.fdLog("fdCopySelectionStandard - cannot copy text using API");
    return {
      success: false,
      errorMessage: 'cannot copy text using API',
      supported: true,
      sendKeysFromFD: true
    };
  }

  window.fdPasteSelectionStandard = function (textControlEl) {
    window.fdLog("fdPasteSelectionStandard - controlId=" + textControlEl.fdUniqueId);

    var mainWindow = window.fdGetMainWindow();
    if (mainWindow.document.execCommand('paste', false, null)) {
      return {
        success: true,
        supported: true,
        sendKeysFromFD: false
      };
    }

    window.fdLog("fdPasteSelectionStandard - cannot paste text using API");
    return {
      success: false,
      errorMessage: 'cannot paste text using API',
      supported: true,
      sendKeysFromFD: true
    };
  }

  window.fdScrollUpStandard = function (textControlEl) {
    window.fdLog("fdScrollUpStandard - controlId=" + textControlEl.fdUniqueId);
    textControlEl.scrollTop = textControlEl.scrollTop - (textControlEl.clientHeight / 2);
    return {
      success: true,
      supported: true,
      sendKeysFromFD: false
    };
  }

  window.fdScrollDownStandard = function (textControlEl) {
    window.fdLog("fdScrollDownStandard - controlId=" + textControlEl.fdUniqueId);
    textControlEl.scrollTop = textControlEl.scrollTop + (textControlEl.clientHeight / 2);
    return {
      success: true,
      supported: true,
      sendKeysFromFD: false
    };
  }

  window.fdScrollCaretStandard = function (textControlEl) {
    window.fdLog("fdScrollCaretStandard - controlId=" + textControlEl.fdUniqueId);
    window.fdLog("fdScrollCaretStandard - cannot scroll to caret using API");

    return {
      success: false,
      errorMessage: 'cannot scroll to caret using API',
      supported: false,
      sendKeysFromFD: false
    };
  }

  window.fdUndoStandard = function (textControlEl) {
    window.fdLog("fdUndoStandard - controlId=" + textControlEl.fdUniqueId);
    window.fdLog("fdUndoStandard - cannot undo using API");

    return {
      success: false,
      errorMessage: 'cannot undo using API',
      supported: true,
      sendKeysFromFD: true
    };
  }
}

// Meditech w/ M*Modal API
// Create FD text control functions that wraps Meditech's MSpeech API.
if (window.MSpeechAPI) {
  window.fdLog('fdWebExtension - found Meditech');

  if (!window.fdInitializeCustomApi) {
    window.fdCustomApiInitialized = false;

    window.fdInitializeCustomApi = function () {
      if (window.fdCustomApiInitialized) {
        return { success: true };
      }

      window.fdLog('fdInitializeCustomApi - initializing MSpeechAPI');

      var initResult = window.MSpeechAPI('initialize()');
      window.fdCustomApiInitialized = initResult[0];
      if (window.fdCustomApiInitialized) {
        window.fdLog("fdInitializeCustomApi - MSpeechAPI initialized");
        return { success: true };
      }
      else {
        window.fdLog("fdInitializeCustomApi - MSpeechAPI could not be initialized (reason='" + initResult[1] + "')");
        return {
          success: window.fdCustomApiInitialized,
          errorMessage: initResult[1]
        };
      }
    }
  }

  if (!window.fdGetFocusedControlIdCustom) {
    window.fdGetFocusedControlIdCustom = function () {
      window.fdLog('fdGetFocusedControlIdCustom - getting focused control');

      if (!window.fdInitializeCustomApi().success) {
        window.fdLog('fdGetFocusedControlIdCustom - MSpeechAPI is not initialized');
        return {
          success: false,
          errorMessage: 'MSpeechAPI is not initialized',
          focusedControlId: null
        };
      }

      var focusedControlId = window.MSpeechAPI('controls.getCurrentControl()')[0];
      window.fdLog("fdGetFocusedControlIdCustom - controlId=" + focusedControlId);
      return {
        success: true,
        focusedControlId: focusedControlId
      };
    }
  }

  if (!window.fdFindControlCustom) {
    window.fdFindControlCustom = function (controlId) {
      window.fdLog('fdFindControlCustom - controlId=' + controlId);

      var controlIds = window.MSpeechAPI('controls.enumerate()');
      for (var i = 0; i < controlIds.length; i++) {
        if (controlIds[i].localeCompare(controlId) == 0) {
          window.fdLog('fdFindControlCustom - found controlId=' + controlId);
          return true;
        }
      }

      window.fdLog('fdFindControlCustom - did not find controlId=' + controlId);

      return false;
    }
  }

  if (!window.fdGetTextCustom) {
    window.fdGetTextCustom = function (controlId) {
      window.fdLog('fdGetTextCustom - controlId=' + controlId);
      var text = window.MSpeechAPI('control.text.read(id)', controlId)[0];
      window.fdLog("fdGetTextCustom - text='" + text.replace(/\n/g, '\\n') + "'");
      return {
        success: true,
        text: text
      };
    }
  }

  if (!window.fdGetTextLengthCustom) {
    window.fdGetTextLengthCustom = function (controlId) {
      window.fdLog('fdGetTextLengthCustom - controlId=' + controlId);
      var text = window.fdGetTextCustom(controlId).text;
      var textLength = text ? text.length : 0;
      window.fdLog('fdGetTextLengthCustom - textLength=' + textLength);
      return {
        success: true,
        textLength: textLength
      };
    }
  }

  if (!window.fdGetSelectionCustom) {
    window.fdGetSelectionCustom = function (controlId) {
      window.fdLog('fdGetSelectionCustom - controlId = ' + controlId);
      var sel = window.MSpeechAPI('control.selection.getRange(id)', controlId);
      window.fdLog('fdGetSelectionCustom - selStart=' + sel[0] + ' selEnd=' + (sel[0] + sel[1]) + ' selLength=' + sel[1]);
      return {
        success: true,
        start: sel[0],
        end: sel[0] + sel[1],
        length: sel[1]
      }
    }
  }

  if (!window.fdSetSelectionCustom) {
    window.fdSetSelectionCustom = function (controlId, selStart, selEnd) {
      window.fdLog('fdSetSelectionCustom - controlId=' + controlId + ' selStart=' + selStart + ' selEnd' + selEnd);
      if (selEnd >= 0) {
        window.MSpeechAPI('control.selection.setRange(id,point#,extent#)', controlId, selStart, selEnd - selStart);
      }
      else {
        window.fdLog('fdSetSelectionCustom - select all text');
        var text = window.fdGetText(controlId).text;
        if (text) {
          window.MSpeechAPI('control.selection.setRange(id,point#,extent#)', controlId, 0, text.length);
        }
      }

      return { success: true };
    }
  }

  if (!window.fdInsertTextCustom) {
    window.fdInsertTextCustom = function (controlId, text) {
      window.fdLog("fdInsertTextCustom - controlId=" + controlId + " text='" + text.replace(/\n/g, '\\n') + "'");
      var sel = window.fdGetSelectionCustom(controlId);
      try {
        window.MSpeechAPI('control.text.write(id,point#,extent#,text)', controlId, sel.start, sel.length, text);
      }
      catch (e) {
        // Meditech sample page is throwing an exception, but it seems to be after text has already been inserted
        // so just catch the exception, log it, and continue on
        console.error(e);
      }
      window.fdSetSelectionCustom(controlId, sel.start + text.length, sel.start + text.length);

      // simulate fake key down event for search fields
      // this isn't needed outside of search fields, but FD cannot tell if it's in a regular field or a search field
      var mainWindow = window.fdGetMainWindow();
      var activeElement = mainWindow.document.activeElement;
      if (activeElement) {
        try {
          var document = mainWindow.document;
          var eventObj = document.createEventObject ? document.createEventObject() : document.createEvent("Events");
          if (eventObj.initEvent) {
            eventObj.initEvent("keydown", true, true);
          }
          activeElement.dispatchEvent ? activeElement.dispatchEvent(eventObj) : activeElement.fireEvent("onkeydown", eventObj);
        }
        catch (e) { }
      }

      return { success: true };
    }
  }

  if (!window.fdDeleteSelectionCustom) {
    window.fdDeleteSelectionCustom = function (controlId) {
      window.fdLog("fdDeleteSelectionCustom - controlId=" + controlId);
      return window.fdInsertTextCustom(controlId, '');
    }
  }

  if (!window.fdBoldSelectionCustom) {
    window.fdBoldSelectionCustom = function (controlId) {
      window.fdLog("fdBoldSelectionCustom - controlId=" + controlId);
      return {
        success: false,
        errorMessage: 'cannot bold text using MSpeechAPI',
        supported: true,
        sendKeysFromFD: true
      };
    }
  }

  if (!window.fdItalicizeSelectionCustom) {
    window.fdItalicizeSelectionCustom = function (controlId) {
      window.fdLog("fdItalicizeSelectionCustom - controlId=" + controlId);
      return {
        success: false,
        errorMessage: 'cannot italicize text using MSpeechAPI',
        supported: true,
        sendKeysFromFD: true
      };
    }
  }

  if (!window.fdUnderlineSelectionCustom) {
    window.fdUnderlineSelectionCustom = function (controlId) {
      window.fdLog("fdUnderlineSelectionCustom - controlId=" + controlId);
      return {
        success: false,
        errorMessage: 'cannot underline text using MSpeechAPI',
        supported: true,
        sendKeysFromFD: true
      };
    }
  }

  if (!window.fdCutSelectionCustom) {
    window.fdCutSelectionCustom = function (controlId) {
      window.fdLog("fdCutSelectionCustom - controlId=" + controlId);
      return {
        success: false,
        errorMessage: 'cannot cut text using MSpeechAPI',
        supported: true,
        sendKeysFromFD: true
      };
    }
  }

  if (!window.fdCopySelectionCustom) {
    window.fdCopySelectionCustom = function (controlId) {
      window.fdLog("fdCopySelectionCustom - controlId=" + controlId);
      return {
        success: false,
        errorMessage: 'cannot copy text using MSpeechAPI',
        supported: true,
        sendKeysFromFD: true
      };
    }
  }

  if (!window.fdPasteSelectionCustom) {
    window.fdPasteSelectionCustom = function (controlId) {
      window.fdLog("fdPasteSelectionCustom - controlId=" + controlId);
      return {
        success: false,
        errorMessage: 'cannot paste text using MSpeechAPI',
        supported: true,
        sendKeysFromFD: true
      };
    }
  }

  if (!window.fdScrollUpCustom) {
    window.fdScrollUpCustom = function (controlId) {
      window.fdLog("fdScrollUpCustom - controlId=" + controlId);
      return {
        success: false,
        errorMessage: 'cannot scroll up using MSpeechAPI',
        supported: true,
        sendKeysFromFD: true
      };
    }
  }

  if (!window.fdScrollDownCustom) {
    window.fdScrollDownCustom = function (controlId) {
      window.fdLog("fdScrollDownCustom - controlId=" + controlId);
      return {
        success: false,
        errorMessage: 'cannot scroll down using MSpeechAPI',
        supported: true,
        sendKeysFromFD: true
      };
    }
  }

  if (!window.fdScrollCaretCustom) {
    window.fdScrollCaretCustom = function (controlId) {
      window.fdLog("fdScrollCaretCustom - controlId=" + controlId);
      var sel = window.fdGetSelection(controlId);
      window.fdSetSelection(controlId, sel.start, sel.start + sel.length);
      return {
        success: true,
        supported: true,
        sendKeysFromFD: false
      };
    }
  }

  if (!window.fdUndoCustom) {
    window.fdUndoCustom = function (controlId) {
      window.fdLog("fdUndoCustom - controlId=" + controlId);
      return {
        success: false,
        errorMessage: 'cannot undo using MSpeechAPI',
        supported: !window.fdIsSamplePage(),  // undo doesn't work on sample page
        sendKeysFromFD: true
      };
    }
  }

  if (!window.fdIsSamplePage) {
    window.fdIsSamplePage = function () {
      return window.document.title.indexOf('MSpeech Sample') >= 0;
    }
  }
}

if (!window.fdInitializeApi) {
  window.fdInitializeApi = function () {
    var customResult = window.fdInitializeCustomApi ? window.fdInitializeCustomApi() : null;
    if (customResult && !customResult.success) {
      return customResult;
    }

    return { success: true };
  }

  window.fdGetFocusedControlId = function () {
    var customResult = window.fdGetFocusedControlIdCustom ? window.fdGetFocusedControlIdCustom() : null;
    if (customResult && (customResult.focusedControlId || customResult.errorMessage)) {
      return customResult;
    }

    return window.fdGetFocusedControlIdStandard();
  }

  window.fdGetText = function (controlId) {
    window.fdLog('fdGetText - controlId=' + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdGetText - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified',
        text: null
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdGetTextCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdGetTextStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found',
      text: null
    }
  }

  window.fdGetTextLength = function (controlId) {
    window.fdLog('fdGetTextLength - controlId=' + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdGetTextLength - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified',
        textLength: 0
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdGetTextLengthCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdGetTextLengthStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found',
      textLength: 0
    }
  }

  window.fdGetSelection = function (controlId) {
    window.fdLog('fdGetSelection - controlId=' + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdGetSelection - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified',
        start: 0, end: 0, length: 0
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdGetSelectionCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdGetSelectionStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found',
      start: 0, end: 0, length: 0
    }
  }

  window.fdSetSelection = function (controlId, selStart, selEnd) {
    window.fdLog('fdSetSelectionCustom - controlId=' + controlId + ' selStart=' + selStart + ' selEnd' + selEnd);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdSetSelection - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdSetSelectionCustom(controlId, selStart, selEnd);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdSetSelectionStandard(textControlEl, selStart, selEnd);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdInsertText = function (controlId, text) {
    window.fdLog("fdInsertText - controlId=" + controlId + " text='" + text.replace(/\n/g, '\\n') + "'");

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdInsertText - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdInsertTextCustom(controlId, text);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdInsertTextStandard(textControlEl, text);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdDeleteSelection = function (controlId, text) {
    window.fdLog("fdDeleteSelection - controlId=" + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdDeleteSelection - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdDeleteSelectionCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdDeleteSelectionStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdBoldSelection = function (controlId) {
    window.fdLog("fdBoldSelection - controlId=" + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdBoldSelection - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdBoldSelectionCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdBoldSelectionStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdItalicizeSelection = function (controlId) {
    window.fdLog("fdItalicizeSelection - controlId=" + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdItalicizeSelection - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdItalicizeSelectionCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdItalicizeSelectionStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdUnderlineSelection = function (controlId) {
    window.fdLog("fdUnderlineSelection - controlId=" + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdUnderlineSelection - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdUnderlineSelectionCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdUnderlineSelectionStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdCutSelection = function (controlId) {
    window.fdLog("fdCutSelection - controlId=" + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdCutSelection - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdCutSelectionCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdCutSelectionStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdCopySelection = function (controlId) {
    window.fdLog("fdCopySelection - controlId=" + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdCopySelection - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdCopySelectionCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdCopySelectionStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdPasteSelection = function (controlId) {
    window.fdLog("fdPasteSelection - controlId=" + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdPasteSelection - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdPasteSelectionCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdPasteSelectionStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdScrollUp = function (controlId) {
    window.fdLog("fdScrollUp - controlId=" + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdScrollUp - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdScrollUpCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdScrollUpStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdScrollDown = function (controlId) {
    window.fdLog("fdScrollDown - controlId=" + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdScrollDown - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdScrollDownCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdScrollDownStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdScrollCaret = function (controlId) {
    window.fdLog("fdScrollCaret - controlId=" + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdScrollCaret - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdScrollCaretCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdScrollCaretStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdUndo = function (controlId) {
    window.fdLog("fdUndo - controlId=" + controlId);

    var focusedControlId = window.fdGetFocusedControlId().focusedControlId;
    if (focusedControlId != controlId) {
      controlId = focusedControlId;
      window.fdLog('fdUndo - focused control Id changed - controlId=' + controlId);
    }

    if (!controlId) {
      return {
        success: false,
        errorMessage: 'controlId must be specified'
      }
    }

    if (window.fdFindControlCustom && window.fdFindControlCustom(controlId)) {
      return window.fdUndoCustom(controlId);
    }

    var textControlEl = window.fdFindControlStandard(controlId);
    if (textControlEl) {
      return window.fdUndoStandard(textControlEl);
    }

    return {
      success: false,
      errorMessage: 'text control not found'
    }
  }

  window.fdAddMessageHandler('TextControlHandler', function (obj) {
    if (obj.msg == 'GetFocusedControlId') {
      var result = window.fdGetFocusedControlId();
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          focusedControlId: result.focusedControlId
        }
      }));

      return true;
    }
    else if (obj.msg == 'GetText') {
      var result = window.fdGetText(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          text: result.text
        }
      }));

      return true;
    }
    else if (obj.msg == 'GetTextLength') {
      var result = window.fdGetTextLength(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          textLength: result.textLength
        }
      }));

      return true;
    }
    else if (obj.msg == 'GetSelection') {
      var result = window.fdGetSelection(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          selectionStart: result.start,
          selectionEnd: result.end
        }
      }));

      return true;
    }
    else if (obj.msg == 'SetSelection') {
      var result = window.fdSetSelection(obj.controlId, obj.selectionStart, obj.selectionEnd);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
        }
      }));

      return true;
    }
    else if (obj.msg == 'DeleteSelection') {
      var result = window.fdDeleteSelection(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
        }
      }));

      return true;
    }
    else if (obj.msg == 'InsertText') {
      var result = window.fdInsertText(obj.controlId, obj.text);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
        }
      }));

      return true;
    }
    else if (obj.msg == 'BoldSelection') {
      var result = window.fdBoldSelection(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          supported: result.supported,
          sendKeysFromFD: result.sendKeysFromFD
        }
      }));

      return true;
    }
    else if (obj.msg == 'ItalicizeSelection') {
      var result = window.fdItalicizeSelection(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          supported: result.supported,
          sendKeysFromFD: result.sendKeysFromFD
        }
      }));

      return true;
    }
    else if (obj.msg == 'UnderlineSelection') {
      var result = window.fdUnderlineSelection(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          supported: result.supported,
          sendKeysFromFD: result.sendKeysFromFD
        }
      }));

      return true;
    }
    else if (obj.msg == 'CutSelection') {
      var result = window.fdCutSelection(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          supported: result.supported,
          sendKeysFromFD: result.sendKeysFromFD
        }
      }));

      return true;
    }
    else if (obj.msg == 'CopySelection') {
      var result = window.fdCopySelection(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          supported: result.supported,
          sendKeysFromFD: result.sendKeysFromFD
        }
      }));

      return true;
    }
    else if (obj.msg == 'PasteSelection') {
      var result = window.fdPasteSelection(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          supported: result.supported,
          sendKeysFromFD: result.sendKeysFromFD
        }
      }));

      return true;
    }
    else if (obj.msg == 'ScrollUp') {
      var result = window.fdScrollUp(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          supported: result.supported,
          sendKeysFromFD: result.sendKeysFromFD
        }
      }));

      return true;
    }
    else if (obj.msg == 'ScrollDown') {
      var result = window.fdScrollDown(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          supported: result.supported,
          sendKeysFromFD: result.sendKeysFromFD
        }
      }));

      return true;
    }
    else if (obj.msg == 'ScrollCaret') {
      var result = window.fdScrollCaret(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          supported: result.supported,
          sendKeysFromFD: result.sendKeysFromFD
        }
      }));

      return true;
    }
    else if (obj.msg == 'Undo' || obj.msg == 'UndoSelection') {
      var result = window.fdUndo(obj.controlId);
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: result.success,
          errorMessage: result.errorMessage ? result.errorMessage : '',
          supported: result.supported,
          sendKeysFromFD: result.sendKeysFromFD
        }
      }));

      return true;
    }

    return false;
  });

  window.fdInitializeApi();
  window.fdConnect();
}

// Any EMR SSO handler
// If fdGetUserId function is defined, then add a message handler for the GetUserId messages.
if (window.fdGetUserId) {
  window.fdLog("fdWebExtension - found fdGetUserId SSO function");

  window.fdAddMessageHandler('SsoHandler', function (obj) {
    if (obj.msg == 'GetUserId') {
      window.fdLog('fdSsoHandler - getting userId');
      var userId = window.fdGetUserId();
      window.fdLog("fdSsoHandler - userId='" + userId + "'");
      window.fdWebSocket.send(JSON.stringify({
        id: obj.id,
        msg: obj.msg,
        result: {
          success: true,
          userId: userId
        }
      }));
      return true;
    }

    return false;
  });

  window.fdConnect();
}

// Any EMR grammar handler
// If FD grammar functions are defined, then set up a timer and poll for grammar changes.
// If grammar changes, then notify FD of the grammar via a web socket message.
// If FD recognizes a command from the grammar, then it'll send a message to be handled by fdExecuteCommand.
if (window.fdGetGrammar && window.fdExecuteCommand) {
  window.fdLog('fdWebExtension - found grammar functions');

  if (!window.fdGrammarTimerId) {
    var currentGrammar = '';

    window.fdAddMessageHandler('GrammarHandler', function (obj) {
      if (obj.msg == 'ExecuteCommand') {
        window.fdLog('fdGrammarHandler - execute command resultDetail=' + obj.resultDetailJson);
        var resultDetailObj = JSON.parse(obj.resultDetailJson);
        window.fdExecuteCommand(resultDetailObj);
        return true;
      }

      return false;
    });

    window.fdConnect();

    window.fdGrammarTimerId = window.setInterval(function () {

      // if web socket isn't ready yet, then set current grammar to empty
      // as soon as web socket is ready, then grammar change is detected
      // this is useful for when FD hasn't started yet or is restarted
      if (!window.fdWebSocket || window.fdWebSocket.readyState != 1) {
        currentGrammar = '';
        return;
      }

      var grammar = window.fdGetGrammar();
      if (grammar != currentGrammar) {
        currentGrammar = grammar;

        if (currentGrammar) {
          window.fdLog('fdGrammarTimer - newGrammar=' + currentGrammar);

          window.fdWebSocket.send(JSON.stringify({
            id: window.fdGuid(),
            msg: 'AddGrammar',
            fieldId: 'AnyEmrGrammar',
            grammar: currentGrammar
          }));
        }
        else {
          window.fdLog('fdGrammarTimer - grammar is empty');

          window.fdWebSocket.send(JSON.stringify({
            id: window.fdGuid(),
            msg: 'RemoveGrammar',
            fieldId: 'AnyEmrGrammar'
          }));
        }
      }
    }, 1000);
  }
}